#include <stdio.h>
#include "cv.h"
#include "cxcore.h"
#include "highgui.h"
#include "math.h"
#include "perftimer.h"

#define OPT_NOSEP
//#define OPT_SEP

#define PIXVAL(iplimagep, x, y) (*(uchar *)((iplimagep)->imageData + (y) * (iplimagep)->widthStep + (x)))

#define SIGMA 5
#define MARGIN 2
#define MCENTER (MARGIN + 1)
#define MSIZE (MARGIN + MARGIN + 1)
double w1[MSIZE];
double w2[MSIZE * MSIZE];

#define CLIP_MAX 255
#define CLIP_MIN 0

uchar clipvalue(double v)
{
    int vi = (int)v;
    if (vi >= CLIP_MAX) {
        return CLIP_MAX;
    } else if (vi <= CLIP_MIN) {
        return CLIP_MIN;
    } else {
        return (uchar)vi;
    }
}

perftimer *pt;

int main()
{
    int i, j, m, n;

    for (i = 0; i < MSIZE; i++) {
	w1[i] = 1.0 / sqrt(2 * CV_PI) / SIGMA
	    * exp(- (i - MCENTER) * (i - MCENTER) / (2.0 * SIGMA * SIGMA));
    }
    for (j = 0; j < MSIZE; j++) {
	for (i = 0; i < MSIZE; i++) {
	    w2[MSIZE * j + i] = w1[i] * w1[j];
	}
    }
    pt = create_perftimer(100, 1);

    CvCapture *capture = cvCaptureFromCAM(0);

    IplImage *frame, *buf, *mono, *mid, *result;

    if (!capture) {
	fprintf(stderr, "cannot capture\n");
	return 1;
    }
    frame = cvQueryFrame(capture);
    buf = cvCreateImage(cvGetSize(frame), IPL_DEPTH_8U, 3);  // color
    mono = cvCreateImage(cvGetSize(frame), IPL_DEPTH_8U, 1);
    mid = cvCreateImage(cvGetSize(frame), IPL_DEPTH_8U, 1);
    result = cvCreateImage(cvGetSize(frame), IPL_DEPTH_8U, 1);

    cvNamedWindow("input", CV_WINDOW_AUTOSIZE);
    cvNamedWindow("result", CV_WINDOW_AUTOSIZE);

    while (1) {
	frame = cvQueryFrame(capture); /* cannot be modified */

	cvFlip(frame, buf, NULL);
	cvCvtColor(buf, mono, CV_BGR2GRAY);

	start_perftimer(pt, 0);
#if defined(OPT_NOSEP)
	for (j = MARGIN; j < mono->height - MARGIN; j++) {
	    for (i = MARGIN; i < mono->width - MARGIN; i++) {
		double sum = 0.0;
		for (n = 0; n < MSIZE; n++) {
		    for (m = 0; m < MSIZE; m++) {
			sum += w2[MSIZE * n + m]
			    * PIXVAL(mono, i - MARGIN + m, j - MARGIN + n);
		    }
		}
		PIXVAL(result, i, j) = clipvalue(sum);
	    }
	}
#elif defined(OPT_SEP)
	for (j = MARGIN; j < mono->height - MARGIN; j++) {
	    for (i = MARGIN; i < mono->width - MARGIN; i++) {
		double sum = 0.0;
		for (m = 0; m < MSIZE; m++) {
		    sum += w1[m] * PIXVAL(mono, i - MARGIN + m, j);
		}
		PIXVAL(mid, i, j) = clipvalue(sum);
	    }
	}
	for (j = MARGIN; j < mono->height - MARGIN; j++) {
	    for (i = MARGIN; i < mono->width - MARGIN; i++) {
		double sum = 0.0;
		for (m = 0; m < MSIZE; m++) {
		    sum += w1[m] * PIXVAL(mid, i, j - MARGIN + m);
		}
		PIXVAL(result, i, j) = clipvalue(sum);
	    }
	}
#endif
	stop_perftimer(pt, 0);
	
	cvShowImage("input", mono);
	cvShowImage("result", result);

	if (cvWaitKey(30) > 0) {
	    break;
	}
    }

    print_stat_perftimer(pt, stdout);
    
    return 0;
}
